#!/bin/sh
# \
	exec bp_tcl "$0" ${1+"$@"}


#------------------------------------------------------
#
# splitTrain - Load  Data from file and split it
# so that section from [i/n i+1/n] goes into validation
# and remainder into test
#
#------------------------------------------------------
proc splitData { file i {n 4} } {
global numInputs numTargets

   set fd [open "|grep -c ^ $file"]
   gets $fd num
   close $fd;
   set minv [expr int($i * $num / $n)];
   set maxv [expr int( ($i + 1) * $num / $n)];

   set fdread [open $file r]
   set fdtrain [open train w];
   set fdvalid [open valid w];


   set line 0;
   while { [gets $fdread data] >= 0} {
      if ![info exists numInputs] {
	 set numTargets 1 ;
	 set numInputs	[expr [llength $data] -  $numTargets] ;
      }	

      if { $line >= $minv && $line < $maxv} {
	 puts $fdvalid $data ;
      } else {
	 puts $fdtrain $data
      }
      incr line ;

   }

   close $fdread
   close $fdvalid
   close $fdtrain

}

proc addSplitData {} {
global trainSet validSet

   foreach set [uts_info exampleSets] {
      $set delete
   }

   uts_exampleSet $trainSet
   uts_exampleSet $validSet

   uts_addExamples net $trainSet train
   uts_addExamples net $validSet valid
}

# ----------------------------------------------------------------------
# Estimate the number of hidden units to use. The method estimates the
# amount of information in the training set $i as 0.5 log(var_r / var)
# with var_r variance of the residual of prediction of a linear
# model to the variance of the training set targets (This assumes
# only a single target) Note must use base 2 to get I in bits
# 
# The number of hiddens is given by assuming r bits/weight so
#
#    r * (nh (ni + 2) + 1) = N I
#
# N - number of training examples
# I - Information in bits as estimated above
# ni - Number of input units
# nh - Number of hidden units
#
# Note includes bias weights on the hidden and output unit
# Uses the linear model fitting program lin
# ----------------------------------------------------------------------
proc estHidden { ni r } {
#
# ni - Number of input units
# r  - Allow this many bits per weight
# i  - Training set number

   set fid [open "|lin -p tmp1 -t train.0 train.0"]
   set line [read $fid]
   close $fid

   # The lin program gives a single line report
   # The last element is the variance of the targets
   # The third last element is variance of the residual
   # The 7'th element is the number of training examples
   set len [llength $line]
   set rat [expr [lindex $line [expr $len - 1] ] / [lindex $line [expr $len - 3]]]
   set N [lindex $line 6];

   # Information in bits
   set I [expr log($rat) / log(2) / 2] ;

   # Estimate number of hidden units min of 1 hidden unit
   set nh [expr round(($N * $I / $r - 1) / ($ni + 2))]
   set nh [ expr $nh >  0 ? $nh : 1];

   return $nh
}

#------------------------------------------------------
#
#  trainSplit - Train a network on all splits s of task 
#   instance  t
#
# -----------------------------------------------------
proc trainSplit {file} {
global wcList bitsList nSplit mz mzIter
global numInputs numTargets
global trainSet validSet
global trainArr validArr numHidden
global  dataset prototask task progFile


   loop j 0 $nSplit {
      splitData $file $j $nSplit
      set net [buildRegress $numInputs 0 $numTargets]
      addSplitData

      foreach bitPerWeight $bitsList {
	 set numHidden($bitPerWeight) [estHidden $numInputs $bitPerWeight ] 

	 set net [buildRegress $numInputs $numHidden($bitPerWeight) $numTargets]
	 $mz configure -net $net -exampleSet $trainSet

	 foreach wc $wcList {
	    set fd [open "|date" r]
	    gets $fd date
	    close $fd

	   r  puts "Starting at $date on $dataset/$prototask/$task $file split: $j  bits: $bitPerWeight WC: $wc" >> ../../../$progFile

	    net.costModel configure -scale $wc
	    uts_randomizeNet $net
	    $mz reset
	    r $mz run -iterations $mzIter > /dev/null
	    set failStatus [lindex [lindex [$mz status ] 6] 1]

	    if { 0 != $failStatus} {
	       set res [reportTrainValidPerf $j $numHidden($bitPerWeight) $wc]
	       set trainArr($j,$bitPerWeight,$wc) [lindex $res 0];
	       set validArr($j,$bitPerWeight,$wc) [lindex $res 1];
	    } else {
	       print [lindex [lindex [$mz status ] 7] 1]
	    }	
	    set sta [$mz status]
	    r puts "[lindex $sta 1] [lindex $sta 2]" >> ../../../$progFile
	 }
      }
   }

   return [bestConfig validArr numHidden] 
}

#-------------------------------------------------------
# 
# buildRegress - A regresion net with no direct connections
#
#-------------------------------------------------------

proc buildRegress {numInput numHidden numOutput {outputType LINEAR}} {

   foreach n [uts_info nets] {
      $n delete
   }

   set net [uts_basicNet net] ;

   if {$numHidden > 0 } {
       uts_addGroups $net  "Input $numInput INPUT" \
	 "Hidden $numHidden {}" "Output $numOutput {OUTPUT $outputType}" ;

      uts_connectNet $net {Input Hidden} {Hidden Output} {Input Output}
   } else {

      uts_addGroups $net  "Input $numInput INPUT" "Output $numOutput {OUTPUT $outputType}" ;
       uts_connectNet $net {Input Output};
    }

   # Delete the bias from the cost Model
   foreach link [getGroupOutgoingWeights net.Bias] {
      net.costModel deleteLink [lindex $link 0] [lindex $link 1]
   }

   uts_set uts_net<0>.extension.zeroErrorRadius 0.0


   return $net
}


#------------------------------------------------------
#
# returns
#         0.5 sum (y - yh)^2   / N
# for the train and validation sets
#------------------------------------------------------
proc reportTrainValidPerf { splitN nHid wc } {
global trainSet validSet  exampleTrace mz

   if {[info exists exampleTrace]} {
      set exampleTraceSave $exampleTrace
      unset exampleTrace
   }

   uts_doExamples net $trainSet
   set trainErr [expr [lindex [net configure -error] 2] / [$trainSet size]]
   uts_doExamples net $validSet
   set validErr [expr [lindex [net configure -error] 2] / [$validSet size]]
   set result [$mz status] ;

# Report NumHidden Cg-Iter NFe trainErr validErr	    
#   print $splitN $nHid $wc [lindex [lindex $result 1] 1] $trainErr $validErr

   if {[info exists exampleTraceSave]} {
      set exampleTrace $exampleTraceSave
   }

   return [list $trainErr $validErr]
}


proc reportResults { what best a  } {
global wcList bitsList nSplit numHidden
global method dataset prototask task

   upvar $a table

   puts "Task: $method/$dataset$prototask/$task";
   puts "Selected WC: [lindex $best 1]  NumHidden [lindex $best 3]\n"

   foreach bitPerWeight $bitsList {
   
      puts "Train Data: $what\nBitsPer weight: $bitPerWeight  Hidden Units $numHidden($bitPerWeight)"

      puts "WC\t\t0\t\t1\t\t2\t\t3\t  Total"
      foreach wc $wcList {
	 set str "$wc"
	 set tot 0
	 loop j 0 $nSplit {	 
	    set tot [expr $tot + $table($j,$bitPerWeight,$wc)];
	    set str "$str [format " \t%9.3f" $table($j,$bitPerWeight,$wc)]"
	 }
	 puts "$str[format " \t%9.3f" $tot]"
      }
   puts "\n"
   }

}

proc bestConfig {res hid} {
global wcList bitsList nSplit

   upvar $res table
   upvar $hid numHid

   foreach bitPerWeight $bitsList {
      foreach wc $wcList {
	 set cost 0;

	 loop j 0  $nSplit {	 
	    set cost [expr $cost + $table($j,$bitPerWeight,$wc)]
	 }

	 if { ![info exists minCost] || $cost < $minCost} {
	    set minCost $cost
	    set bestWc  $wc
	    set bestBit $bitPerWeight
	 }
      }
   }

   # Change sign of minCost
   set minCost [expr -$minCost]
   return [list $minCost $bestWc $bestBit $numHid($bestBit)]
}

proc reportCrossValid {file best} {
global  validArr

   reportResults $file $best validArr 
}

proc addAllData {inst} {
global trainSet validSet

   
   foreach set [uts_info exampleSets] {
      $set delete
   }

   uts_exampleSet $trainSet
   uts_exampleSet $validSet

   uts_addExamples net $trainSet "train.$inst"

   xlateData "test.$inst" "targets.$inst" tmpTest
   uts_addExamples net $validSet tmpTest
   exec rm -f tmpTest
}

# Translate Delve test/target format to Xerion format
proc xlateData { testIn targetsIn testOut } {
global numInputs numTargets

   set fid($testIn)	[open $testIn    r] ;
   set fid($targetsIn)	[open $targetsIn r] ;
   set fid($testOut)	[open $testOut	 w] ;

   while { [gets $fid($testIn) inputs] >= 0
	    && [gets $fid($targetsIn) targets] >= 0} {
      if {![info exists numInputs]} {
	 set numInputs [llength $inputs]
	 set numTargets [llength $targets]
      }
      
	puts $fid($testOut) "$inputs $targets"
   }
   close $fid($testIn) ;
   close $fid($targetsIn) ;
   close $fid($testOut) ;
}

proc trainAll {} {
global trainSet validSet mz mzIter exampleTrace
global numInputs numTargets
global trainArr validArr numHidden 

   set numInstances	[llength [glob -nocomplain {train.[0-9]*}]] ;
   
   loop i 0 $numInstances {
      set file "train.$i"
      set best [trainSplit $file]

      r reportCrossValid $file $best >> log
      
      set wc [lindex $best 1]
      set bestHidden [lindex $best 3]

      set net [buildRegress $numInputs $bestHidden $numTargets]
      $mz configure -net $net -exampleSet $trainSet
      addAllData $i 

      net.costModel configure -scale $wc
      uts_randomizeNet $net
      $mz reset

      r $mz run -iterations $mzIter > /dev/null
      set failStatus [lindex [lindex [$mz status ] 6] 1]
      
      if { 0 != $failStatus} {
	 set exampleTrace {print [printUnit net.Output.0 output]; return 0}
      } else {
	 print [lindex [lindex [$mz status ] 7] 1]
      }	

      r uts_doExamples net $validSet > "cguess.$i"
   }
}

proc uts_addExamples { net exampleSet file {opts all} } {

  if { [uts_info exampleSets $exampleSet] != [list $exampleSet] } {
    error "\"$exampleSet\" is not an exampleSet" ;
  }

  if { [uts_info nets $net] != [list $net] } {
    error "\"$net\" is not a net" ;
  }


  set setSize 2
  set lineNumber 0 ;

  switch -exact $opts \
  "all"  \
  { 
     set nIn [uts_grpTypeSize $net INPUT ]; 
     set nOut [uts_grpTypeSize $net OUTPUT ]; 
  }  \
  "in"  \
  {
     set nIn [uts_grpTypeSize $net INPUT ];
     set nOut 0
  }  \
  "out"  \
   {
     set nOut [uts_grpTypeSize $net OUTPUT ]; 
     set nIn 0
   }

  set inputVector	[uts_vector $nIn] ;
  set targetVector	[uts_vector $nOut] ;

  set nIn1 [max [expr $nIn - 1] 0]
  set nTot [expr $nIn + $nOut]
  set tag {}

  if [catch {
    for_file line $file {
      incr lineNumber ;

      # strip off comments
      set index [string first \# $line] ;
      if { $index >= 0 } {
	set line [string range $line 0 [expr $index - 1]] ;
      }

       if { [llength $line] != $nTot} {
	  error "Expected $nTot items on line $lineNumber but got [llength $line]"
       }

       $exampleSet new $tag $setSize ;

       if {$nIn > 0} {
	  $inputVector list [lrange $line 0 $nIn1]
       }

       if {$nOut > 0} {
	  $targetVector list [lrange $line $nIn end]
       }

       $exampleSet set 0 $inputVector ;
       $exampleSet set 1 $targetVector ;

    }
  } errorMsg] {
    $inputVector delete ;
    $targetVector delete ;

    global errorInfo ;
    error "$errorMsg (before line $lineNumber of $file)" $errorInfo ;
  }
    
  $inputVector delete ;
  $targetVector delete ;
}

set nSplit 4
set bitsList [list 0.5 1 ]
set wcList [list 0 0.008 0.016 0.032 0.064 0.125 0.25 0.5 1 2]
set mzIter 50
set trainSet train
set validSet valid

set uts_readOnly 0

if { [llength $argv] != 5 } {
    puts "usage: [file tail $argv0] method dataset prototask task"
    exit 1 ;
}

set method	[lindex $argv 0] ;
set dataset	[lindex $argv 1] ;
set prototask	[lindex $argv 2] ;
set task	[lindex $argv 3] ;
set progFile   	[lindex $argv 4] ;
puts "Starting $method $dataset $prototask $task"
source $uts_library/utsrc
source $uts_library/bprc

set mz [bp_netMinimizer mz]


trainAll
